from typing import List, Tuple

import torch


def segment_sizes_to_slices(sizes: torch.Tensor) -> List:
    cum_sums = torch.cumsum(sizes, dim=0)
    slices = [slice(0, cum_sums[0])]
    slices += [slice(start, end) for start, end in zip(cum_sums[:-1], cum_sums[1:])]
    return slices


def cut_into_segments(
    data: torch.Tensor, sample_sizes: torch.LongTensor, mask_value: float = -1e9
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Cut data from common tensor into samples part with padding to longest sequence.
    Also return mask with value on pad places.
    :param data: [n contexts; units]
    :param sample_sizes: [batch size]
    :param mask_value:
    :return: [batch size; max context len; units], [batch size; max context len]
    """
    batch_size = len(sample_sizes)
    max_context_len = max(sample_sizes)

    batched_contexts = data.new_zeros((batch_size, max_context_len, data.shape[-1]))
    attention_mask = data.new_zeros((batch_size, max_context_len))

    context_slices = segment_sizes_to_slices(sample_sizes)
    for i, (cur_slice, cur_size) in enumerate(zip(context_slices, sample_sizes)):
        batched_contexts[i, :cur_size] = data[cur_slice]
        attention_mask[i, cur_size:] = mask_value

    return batched_contexts, attention_mask



from typing import List

from tqdm.auto import tqdm


def count_lines_in_file(file_path: str, buffer_size: int = 1024 * 1024) -> int:
    """Count the number of lines in the given file.

    :param file_path: path to the file
    :param buffer_size: size of temporary buffer during reading the file
    :return: number of lines
    """
    n_lines = 0
    with open(file_path, "rb") as file:
        file_reader = file.read
        buffer = file_reader(buffer_size)
        while buffer:
            n_lines += buffer.count(b"\n")
            buffer = file_reader(buffer_size)
    return n_lines


def get_lines_offsets(file_path: str, show_progress_bar: bool = True) -> List[int]:
    """Calculate cumulative offsets for all lines in the given file.

    :param file_path: path to the file
    :param show_progress_bar: if True then tqdm progress bar will be display
    :return: list of ints with cumulative offsets
    """
    line_offsets: List[int] = []
    cumulative_offset = 0
    with open(file_path, "r") as file:
        file_iter = tqdm(file, total=count_lines_in_file(file_path)) if show_progress_bar else file
        for line in file_iter:
            line_offsets.append(cumulative_offset)
            cumulative_offset += len(line.encode(file.encoding))
    return line_offsets


def get_line_by_offset(file_path: str, offset: int) -> str:
    """Get line by byte offset from the given file.

    :param file_path: path to the file
    :param offset: byte offset
    :return: read line
    """
    with open(file_path, "r") as data_file:
        data_file.seek(offset)
        line = data_file.readline().strip()
    return line



from typing import Optional, List

import torch
from torchmetrics import Metric

from commode_utils.metrics import ClassificationMetrics


class SequentialF1Score(Metric):
    def __init__(
        self,
        mask_after_pad: bool = True,
        pad_idx: Optional[int] = None,
        ignore_idx: Optional[List[int]] = None,
        **kwargs,
    ):
        """Initialize the metric for computing f1-score for sequential data.
        This metric is used in many works about code summarization.

        :param mask_after_pad: if True then ignore tokens after first pad index
        :param pad_idx: index of PAD token, required for masking after pad
        :param ignore_idx: indexes that are ignored during computing occurrences
        """
        if mask_after_pad and pad_idx is None:
            raise ValueError("Pass padding index in order to find it first occurrence.")
        super().__init__(**kwargs)
        self._mask_after_pad = mask_after_pad
        self._pad_idx = pad_idx
        self._ignore_idx = ignore_idx if ignore_idx is not None else []

        # metric states
        self.add_state("true_positive", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("false_positive", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("false_negative", default=torch.tensor(0), dist_reduce_fx="sum")

    def _get_after_pad_mask(self, tokens: torch.Tensor) -> torch.Tensor:
        """For each sequence, create mask with all tokens after first PAD

        :param tokens: [seq len; batch size] tensor with tokens to be replaced
        :return: mask with the same shape as tokens
        """
        mask_max_value, mask_max_indices = torch.max(tokens == self._pad_idx, dim=0)
        # if no pad token use len+1 position
        mask_max_indices[~mask_max_value] = tokens.shape[0]
        mask = torch.arange(tokens.shape[0], device=tokens.device).view(-1, 1) >= mask_max_indices
        return mask

    def update(self, predicted: torch.Tensor, target: torch.Tensor):
        """Calculated token occurrence statistic in predicted tensor w.r.t. target tensor.

        :param predicted: [pred seq len; batch size] -- tensor with predicted tokens
        :param target: [target seq len; batch size] -- tensor with ground truth tokens
        :return:
        """
        batch_size = target.shape[1]
        if predicted.shape[1] != batch_size:
            raise ValueError(f"Wrong batch size for prediction (expected: {batch_size}, actual: {predicted.shape[1]})")
        if self._mask_after_pad and self._pad_idx is not None:
            after_pad_mask = self._get_after_pad_mask(predicted)
            predicted[after_pad_mask] = self._pad_idx
            if self._pad_idx not in self._ignore_idx:
                self._ignore_idx.append(self._pad_idx)

        for batch_idx in range(batch_size):
            target_seq = [token for token in target[:, batch_idx] if token not in self._ignore_idx]
            predicted_seq = [token for token in predicted[:, batch_idx] if token not in self._ignore_idx]

            for predicted_token in predicted_seq:
                if predicted_token in target_seq:
                    self.true_positive += 1
                else:
                    self.false_positive += 1
            for target_token in target_seq:
                if target_token not in predicted_seq:
                    self.false_negative += 1

    def compute(self) -> ClassificationMetrics:
        """Calculate precision, recall, and F1-score based on stored token occurrences

        :return: calculated metrics aggregated in data class
        """
        precision = self.true_positive
        if self.true_positive + self.false_positive > 0:
            precision = self.true_positive / (self.true_positive + self.false_positive)
        recall = self.true_positive
        if self.true_positive + self.false_negative > 0:
            recall = self.true_positive / (self.true_positive + self.false_negative)
        f1_score = 2 * precision * recall
        if precision + recall > 0:
            f1_score /= precision + recall
        return ClassificationMetrics(f1_score=f1_score, precision=precision, recall=recall)



from abc import abstractmethod
from os.path import split, join
from typing import Dict, List, Any, Optional

import torch
from pytorch_lightning import Callback, Trainer, LightningModule
from pytorch_lightning.loggers import WandbLogger

from commode_utils.common import print_table


class OldEpochEndCallback(Callback):
    """Implement behaviour of hooks like it was in <=1.1.7 version
    More details:
    https://github.com/PyTorchLightning/pytorch-lightning/issues/5917
    """

    def __init__(self):
        self._is_training = False

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule):
        self._is_training = True

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule):
        self._is_training = False

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule, unused: Optional[bool] = None):
        if trainer.val_dataloaders is None:
            self._old_on_epoch_end(trainer)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        if self._is_training:
            self._old_on_epoch_end(trainer)

    @abstractmethod
    def _old_on_epoch_end(self, trainer: Trainer):
        raise NotImplementedError()


class UploadCheckpointCallback(OldEpochEndCallback):
    """Forced upload saved checkpoint after every epoch. Works only for wandb."""

    def __init__(self, checkpoint_dir: str):
        super().__init__()
        self._checkpoint_dir = checkpoint_dir

    def _old_on_epoch_end(self, trainer: Trainer):
        logger = trainer.logger
        if isinstance(logger, WandbLogger):
            experiment = logger.experiment
            root_dir, _ = split(self._checkpoint_dir)
            experiment.save(join(self._checkpoint_dir, "*.ckpt"), base_path=root_dir)


class PrintEpochResultCallback(OldEpochEndCallback):
    """Print metrics from specified groups after each epoch."""

    def __init__(self, *groups: str, split_symbol: str = "/"):
        super().__init__()
        self._groups = groups
        self._split_symbol = split_symbol

    def _old_on_epoch_end(self, trainer: Trainer):
        metrics_to_print: Dict[str, List[str]] = {group: [] for group in self._groups}
        for key, value in trainer.callback_metrics.items():
            if self._split_symbol not in key:
                continue
            group, metric = key.split(self._split_symbol, 1)
            if group in metrics_to_print:
                if isinstance(value, torch.Tensor):
                    value = value.item()
                metrics_to_print[group].append(f"{metric}={round(value, 2)}")
        metrics_to_print = {k: v for k, v in metrics_to_print.items() if len(v) > 0}
        print_table(metrics_to_print)



import torch
from torch import nn


class LuongAttention(nn.Module):
    """Implement Luong Global Attention with general score function
    https://arxiv.org/pdf/1508.04025.pdf
    """

    def __init__(self, units: int):
        super().__init__()
        self.attn = nn.Linear(units, units, bias=False)

    def forward(self, hidden: torch.Tensor, encoder_outputs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Calculate attention weights
        :param hidden: [batch size; units]
        :param encoder_outputs: [batch size; seq len; units]
        :param mask: [batch size; seq len]
        :return: [batch size; seq len]
        """
        batch_size, seq_len = mask.shape
        # [batch size; units]
        attended_hidden = self.attn(hidden)
        # [batch size; seq len]
        score = torch.bmm(encoder_outputs, attended_hidden.view(batch_size, -1, 1)).squeeze(-1)
        score += mask

        # [batch size; seq len]
        weights = torch.softmax(score, dim=1)
        return weights


class LocalAttention(nn.Module):
    """Simple local attention
    attention_score = W * input_states
    """

    def __init__(self, units: int):
        super().__init__()
        self.attn = nn.Linear(units, 1, bias=False)

    def forward(self, encoder_outputs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Calculate attention weights
        :param encoder_outputs: [batch size; seq len; units]
        :param mask: [batch size; seq len]
        :return: [batch size; 1; seq len]
        """
        # [batch size; seq len; 1]
        attended_encoder_outputs = self.attn(encoder_outputs)
        attended_encoder_outputs += mask.unsqueeze(2)

        # [batch size; seq len; 1]
        weights = torch.softmax(attended_encoder_outputs, dim=1)

        return weights



import subprocess
from os.path import dirname, join
from typing import Dict, List

from omegaconf import DictConfig


def download_dataset(url: str, dataset_dir: str, dataset_name: str):
    download_dir = dirname(dataset_dir)
    download_command_result = subprocess.run(["wget", url, "-P", download_dir], capture_output=True, encoding="utf-8")
    if download_command_result.returncode != 0:
        raise RuntimeError(f"Failed to download dataset. Error:\n{download_command_result.stderr}")
    tar_name = join(download_dir, f"{dataset_name}.tar.gz")
    untar_command_result = subprocess.run(
        ["tar", "-xzvf", tar_name, "-C", download_dir], capture_output=True, encoding="utf-8"
    )
    if untar_command_result.returncode != 0:
        raise RuntimeError(f"Failed to untar dataset. Error:\n{untar_command_result.stderr}")


def print_table(data: Dict[str, List[str]]):
    row_lens = [max(len(header), max([len(s) for s in values])) for header, values in data.items()]
    row_template = " | ".join(["{:<" + str(i) + "}" for i in row_lens])
    headers = [key for key in data.keys()]
    max_data_per_col = max([len(v) for v in data.values()])
    row_data = []
    for i in range(max_data_per_col):
        row_data.append([v[i] if len(v) > i else "" for k, v in data.items()])

    header_line = row_template.format(*headers)
    delimiter_line = "-" * len(header_line)
    row_lines = [row_template.format(*row) for row in row_data]
    print("", header_line, delimiter_line, *row_lines, sep="\n")


def print_config(config: DictConfig, fields: List[str]):
    config_data = {}
    for column in fields:
        if column not in config:
            continue
        config_data[column] = [f"{k}: {v}" for k, v in config[column].items()]
    print_table(config_data)
